/*
Copyright (C) 2005 Parallel Realities

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include "features.h"

void drawFeature(bool transparent)
{
	static Feature *feature;
	static FeatureDef *featureDef;
	static GLColor c;
	
	if (!transparent)
	{
		return;
	}
	
	feature = (Feature*)self;
	featureDef = (FeatureDef*)feature->definition;
	c = featureDef->color;

	switch (feature->featureType)
	{
		case FEATURE_TELEPORTER:
			c.color[3] = (!feature->active) ? 0.25 : 0.5;
			break;
				
		case FEATURE_EXIT:
			c.color[3] = (!feature->active) ? 0.1 : 0.5;
			break;

		case FEATURE_SAVE_POINT:
			c.color[3] = (!feature->active) ? 0.25 : self->custom;
			break;

		case FEATURE_INFORMATION:
			c.color[3] = 0.5;
			break;
	}
	
	if (feature->health > 0)
	{
		drawCyclinder(feature->position, featureDef->segments, featureDef->radius, featureDef->height, c);
	}
	else
	{
		drawCyclinder(feature->position, featureDef->segments, featureDef->radius + (self->health / 10), featureDef->height, c);
	}
}

void fireFeatureTrigger(char *name)
{
	int triggerId = -1;

	for (Feature *feature = (Feature*)entityManager->featureList.getFirstElement() ; feature != NULL ; feature = (Feature*)feature->next)
	{
		if (feature->name == name)
		{
			if (feature->triggerId != -1)
			{
				triggerId = feature->triggerId;
			}

			feature->active = false;
			feature->health = 0;
		}
	}
}

char *extractEnemyFromList(char *enemyList)
{
	static int i;
	static String enemies[25];
	static String list;
	
	i = 0;
	list = enemyList;
	char *token = strtok(list.getText(), "|");
	
	while (token != NULL)
	{
		enemies[i] = token;
		
		token = strtok(NULL, "|");
		
		i++;
	}
	
	return enemies[rand() % i].getText();
}

void spawnerSpawnEntity()
{
	debug(("spawnerSpawnEntity()\n"));
	
	Feature *feature = (Feature*)self;

	feature->thinkTime = Math::rrand(feature->minSpawnTime, feature->maxSpawnTime);
	
	Vector position = feature->position;
	
	Properties props;
	
	props.setProperty("position", position);
	props.setProperty("name", feature->spawnedName.getText());

	if (feature->blobType == "")
	{
		debug(("spawnerSpawnEntity() - Spawning Enemy\n"));
		
		position.x += Math::rrand(-((rand() % 5) * 10), ((rand() % 5) * 10));
		position.y += Math::rrand(-((rand() % 5) * 10), ((rand() % 5) * 10));
		props.setProperty("classname", extractEnemyFromList(feature->enemyType.getText()));
		addEnemy(&props);
	}
	else
	{
		debug(("spawnerSpawnEntity() - Spawning Blob\n"));
		
		props.setProperty("classname", feature->blobType.getText());
		addBlob(&props);
	}

	addTeleportParticles(position);

	audio->playSound(SND_TELEPORT2, CH_SPAWN);

	if ((--feature->custom) == 0)
	{
		feature->health = 0;
		feature->thinkTime = RAND_MAX;
	}
	
	debug(("spawnerSpawnEntity() - Done\n"));
}

void securityPointSearch()
{
	if ((player->health < 0) || (self->health <= 0))
	{
		return;
	}

	if (Math::getDistance(player->position, self->position) <= 10)
	{
		game->setGamePlayMessage(localeManager->translate("HINT_SECURITY"));

		if (game->isControl(CTRL_ACTION))
		{
			game->resetControl(CTRL_ACTION);

			String code;
			code.setText("%.4d", (int)self->custom);

			if (showSecurityDialog(code.getText()))
			{
				self->health = 0;
			}
		}
	}

	self->thinkTime = 1;
}

void exitPointSearch()
{
	if (player->health < 0)
	{
		return;
	}
	
	Feature *feature = (Feature*)self;

	if (Math::getDistance(player->position, self->position) <= 10)
	{
		game->setGamePlayMessage(localeManager->translate("HINT_EXIT"));

		if (game->isControl(CTRL_ACTION))
		{
			if (feature->targetMission != "TITLE_SCREEN")
			{
				game->setMissionStatus(MISSION_COMPLETE);
				game->changingLevel = true;
				game->targetMission = feature->targetMission.getText();
				game->targetLocation = feature->targetLocation.getText();
				game->resetControl(CTRL_ACTION);
			}
			else
			{
				game->setMissionStatus(MISSION_QUITTING);
			}
		}
	}

	self->thinkTime = 1;
}

void teleporterSearch()
{
	addTeleporterParticle(self->position);
	
	for (Entity *ent = bbManager->getFirst(self) ; ent != NULL ; ent = bbManager->getNext())
	{
		if ((ent->entityType != ET_BLOB) && (ent->entityType != ET_ITEM) && (ent->entityType != ET_WEAPON))
		{
			continue;
		}

		if ((ent->health < 0) || (ent->flags & EF_TELEPORTING))
		{
			continue;
		}

		if (Math::getDistance(ent->position, self->position) <= 10)
		{
			addTeleportParticles(ent->position);

			audio->playSound(SND_TELEPORT3, (ent == player) ? CH_PLAYER : CH_SPAWN);

			if (ent->entityType == ET_BLOB)
			{
				ent->destination = self->destination;
				ent->action = teleportUnit;
				ent->velocity.set(0, 0, 0);
				ent->thinkTime = 200;
				ent->contentType = CONTENTS_SOLID;
				((Unit*)ent)->liquidLevel = 0;
	
				Math::addBit(&ent->flags, EF_TELEPORTING);
			}
			else
			{
				ent->position = self->destination;
				addTeleportParticles(ent->position);
			}

			if (ent == player)
			{
				blobsRunToBob();
			}
		}
	}
}

void savePointSearch()
{
	if ((player->health < 0) || (!(player->flags & EF_ONGROUND)))
	{
		return;
	}

	Feature *feature = (Feature*)self;

	feature->thinkTime = 1;

	if (Math::getDistance(feature->position, player->position) <= 10)
	{
		feature->custom += (0.01 * engine->getTimeDifference(TD_LOGIC));
		game->canSaveGame = true;

		game->setGamePlayMessage(localeManager->translate("HINT_SAVE"));

		if (game->isControl(CTRL_ACTION))
		{
			saveGame();
		}
	}
	else
	{
		feature->custom -= (0.01 * engine->getTimeDifference(TD_LOGIC));
	}

	Math::limit(&feature->custom, 0.3, 0.8);
}

void infoPointSearch()
{
	if ((player->health < 0) || (!(player->flags & EF_ONGROUND)))
	{
		return;
	}

	Feature *feature = (Feature*)self;

	feature->thinkTime = 1;

	if (Math::getDistance(feature->position, player->position) <= 6)
	{
		game->setGamePlayMessage(localeManager->translate("HINT_INFO"));

		if (game->isControl(CTRL_ACTION))
		{
			game->resetControl(CTRL_ACTION);
			feature->thinkTime = 5;
			showMessageDialog(localeManager->translate(feature->name.getText()), "SYS_OK");
		}
	}
}

void grapplePointSearch()
{
	if ((player->health < 0) || (!(player->flags & EF_ONGROUND)))
	{
		return;
	}

	Feature *feature = (Feature*)self;

	if (Math::getDistance(feature->position, player->position) <= 10)
	{
		char *neededItem = localeManager->translate("ITM_GRAPPLE_HOOK");
		
		int slot = game->getItemSlot(neededItem);
		
		if ((slot == -1) && (!game->cheatItems))
		{
			game->setGamePlayMessage(localeManager->translate("HINT_NEED_GRAPPLE_HOOK"));
			return;
		}
		
		game->setGamePlayMessage(localeManager->translate("HINT_USE_GRAPPLE_POINT"));
		
		if (game->isControl(CTRL_ACTION))
		{
			game->item[slot]->itemPower--;
			
			if (game->item[slot]->itemPower == 0)
			{
				game->removeItem(neededItem);
			}
			
			game->resetControl(CTRL_ACTION);
			player->position = self->destination;
			
			String s;
			s.setText("str_item_%d", slot);
			textureManager->removeTexture(s.getText());
		}
	}
}

void enableFeature(char *name, bool enable)
{
	for (Feature *feature = (Feature*)entityManager->featureList.getFirstElement() ; feature != NULL ; feature = (Feature*)feature->next)
	{
		if (feature->name == name)
		{
			feature->active = enable;
		}
	}
}

void addFeature(Properties *props)
{
	debug(("addFeature()\n"));
	
	int type = entityManager->getEntityType(props);

	Feature *feature = entityManager->spawnFeature(type);

	switch (type)
	{
		case FEATURE_SPAWNER:
			feature->custom = props->getInt("numberToSpawn", 1);
			feature->action = &spawnerSpawnEntity;
			break;

		case FEATURE_TELEPORTER:
			feature->action = &teleporterSearch;
			break;

		case FEATURE_EXIT:
			feature->action = &exitPointSearch;
			break;

		case FEATURE_SAVE_POINT:
			feature->action = &savePointSearch;
			feature->custom = 0.3;
			break;

		case FEATURE_INFORMATION:
			feature->action = &infoPointSearch;
			feature->custom = 1;
			break;
			
		case FEATURE_GRAPPLE_POINT:
			feature->action = &grapplePointSearch;
			feature->custom = 1;
			break;

		default:
			printf("WARNING: addFeature() - Feature Type not recognised!\n");
			break;
	}
	
	feature->load(props);

	feature->thinkTime = Math::rrand(0, 100);
	feature->draw = drawFeature;
	
	debug(("addFeature() - Done\n"));
}

void doFeatures()
{
	// made true by save points
	game->canSaveGame = false;

	for (Feature *feature = (Feature*)entityManager->featureList.getFirstElement() ; feature != NULL ; feature = (Feature*)feature->next)
	{
		self = feature;
		
		if (feature->health <= 0)
		{
			feature->health -= engine->getTimeDifference(TD_LOGIC);

			if (feature->health <= -100)
			{
				feature = (Feature*)feature->previous;
				entityManager->featureList.remove(feature->next);
			}
		}

		if (feature->active)
		{
			if (feature->performNextAction(engine->getTimeDifference(TD_LOGIC)))
			{
				if (feature->action != NULL)
				{
					feature->action();
				}
			}
		}

		if (withinViewableRange(feature))
		{
			entityManager->addEntityToDrawList(feature);
		}
	}
}
